#!/usr/bin/env python3

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import importlib
import networkx
import queue
import re
import unfurl.parsers
from flask import Flask, render_template, request
from flask_cors import CORS
from unfurl import utils

import logging

log = logging.getLogger(__name__)

# This class and these imports can be removed when they merge https://github.com/MISP/PyMISPWarningLists/pull/15 and
# put out a new PyPI package.
from pymispwarninglists import WarningLists
from pymispwarninglists import api as WarningListsApi
import json
import sys
from glob import glob
from pathlib import Path
from typing import Union, Dict, Any, List, Optional


class FixedWarningLists(WarningLists):
    def __init__(self, slow_search: bool = False, lists: Optional[List] = None):
        """Load all the warning lists from the package.
        :slow_search: If true, uses the most appropriate search method. Can be slower. Default: exact match.
        :lists: A list of warning lists (typically fetched from a MISP instance)
        """

        if not lists:
            lists = []
            self.root_dir_warninglists = Path(
                sys.modules['pymispwarninglists'].__file__).parent / 'data' / 'misp-warninglists' / 'lists'
            for warninglist_file in glob(str(self.root_dir_warninglists / '*' / 'list.json')):
                with open(warninglist_file, mode='r', encoding="utf-8") as f:
                    lists.append(json.load(f))
        if not lists:
            raise api.PyMISPWarningListsError(
                'Unable to load the lists. Do not forget to initialize the submodule (git submodule update --init).')
        self.warninglists = {}
        for warninglist in lists:
            self.warninglists[warninglist['name']] = WarningListsApi.WarningList(warninglist, slow_search)


class Unfurl:
    def __init__(self, remote_lookups=None):
        self.nodes = {}
        self.edges = []
        self.queue = queue.Queue()
        self.next_id = 1
        self.graph = networkx.DiGraph()
        self.total_nodes = 0
        self.api_keys = {}
        self.remote_lookups = remote_lookups
        self.known_domain_lists = None

        config = configparser.ConfigParser()
        config.read('unfurl.ini')
        if config.has_section('API_KEYS'):
            self.api_keys = config['API_KEYS']

        if not self.remote_lookups and config.has_section('UNFURL_APP'):
            self.remote_lookups = config['UNFURL_APP'].getboolean('remote_lookups')

        if not self.known_domain_lists:
            self.build_known_domain_lists()

    class Node:
        def __init__(self, node_id, data_type, key, value, label=None, hover=None,
                     parent_id=None, incoming_edge_config=None, extra_options=None):
            self.node_id = node_id
            self.data_type = data_type
            self.key = key
            self.value = value
            self.label = label
            self.hover = hover
            self.parent_id = parent_id
            self.incoming_edge_config = incoming_edge_config
            self.extra_options = extra_options

            if self.label is None:
                if self.key and self.value:
                    self.label = f'{self.key}: {self.value}'
                elif self.value:
                    self.label = self.value

        def __repr__(self):
            return str(self.__dict__)

    def build_known_domain_lists(self):
        warning_lists = FixedWarningLists()
        warning_lists_dict = warning_lists.warninglists

        # This list has some values I think may confuse users (t.co, drive.google.com, etc), as most things on
        # those domains are not security blog-related, so I'm removing it.
        warning_lists_dict.pop('List of known security providers/vendors blog domain', 1)

        # Similarly, not all bit.ly links are MSOffice-related, so I'm removing it.
        warning_lists_dict['List of known Office 365 URLs'].list.remove('bit.ly')

        # And the capitalization was bothering me, so fixing it here.
        warning_lists_dict['List of known google domains'].name = 'List of known Google domains'
        warning_lists_dict['List of known microsoft domains'].name = 'List of known Microsoft domains'

        self.known_domain_lists = warning_lists_dict

    def search_known_domain_lists(self, domain):
        lists_found_in = []
        for known_list in self.known_domain_lists.values():
            if domain in known_list.list:
                lists_found_in.append({'name': known_list.name, 'description': known_list.description})

        return_list = []
        for found in lists_found_in:
            if not found['name'].startswith('Top'):
                return_list.append(found)

        top_found = [x['name'] for x in lists_found_in if str(x['name']).startswith('Top')]
        top_1k = [x for x in top_found if x.startswith(('Top 1000', 'Top 500 '))]
        top_10k = [x for x in top_found if x.startswith(('Top 10 000', 'Top 10K'))]
        top_1m = [x for x in top_found if x.startswith(('Top 1,000,000', 'Top 20 000'))]

        if top_1k:
            return_list.append({'name': 'Domain is extremely popular (found in "Top 1000" lists)',
                                'description': f'Domain is found in {len(top_1k)} lists: {", ".join(top_1k)}'})

        elif top_10k:
            return_list.append({'name': 'Domain is very popular (found in "Top 10K" lists)',
                                'description': f'Domain is found in {len(top_10k)} lists: {", ".join(top_10k)}'})

        elif top_1m:
            return_list.append({'name': 'Domain is popular (found in "Top 1M" lists)',
                                'description': f'Domain is found in {len(top_1m)} lists: {", ".join(top_1m)}'})

        return return_list

    def get_predecessor_node(self, node):
        if not node.parent_id:
            return False
        predecessor = list(self.graph.predecessors(node))
        return predecessor

    def get_successor_nodes(self, node):
        successors = list(self.graph.successors(node))
        return successors

    def check_sibling_nodes(self, node, data_type=None, key=None, value=None, return_node=False):
        parent_nodes = self.get_predecessor_node(node)

        if not parent_nodes:
            return False

        sibling_nodes = []

        for parent_node in parent_nodes:
            sibling_nodes.extend(self.get_successor_nodes(parent_node))

        for sibling_node in sibling_nodes:

            # Skip the "sibling" if it's actually the source node
            if node.node_id == sibling_node.node_id:
                continue

            # For each attribute, check if it is set. If it is and it
            # doesn't match, stop checking this node and go to the next
            if data_type:
                if data_type != sibling_node.data_type:
                    continue
            if key:
                if key != sibling_node.key:
                    continue
            if value:
                if value != sibling_node.value:
                    continue

            # This node matched all the given criteria;
            if return_node:
                return sibling_node
            return True

        # If we got here, no nodes matched all criteria.
        return False

    def find_preceding_domain(self, node):
        parent_nodes = self.get_predecessor_node(node)

        preceding_domain = ''

        if not parent_nodes:
            return preceding_domain

        assert isinstance(parent_nodes, list)

        for parent_node in parent_nodes:
            if parent_node.data_type == 'url.hostname':
                assert isinstance(parent_node.value, str)
                preceding_domain = parent_node.value
                break
            elif parent_node.data_type == 'url':
                for child_node in self.get_successor_nodes(parent_node):
                    if child_node.data_type == 'url.hostname':
                        assert isinstance(child_node.value, str)
                        preceding_domain = child_node.value
                        break
                    elif child_node.data_type == 'url.authority':
                        for subcomponent in self.get_successor_nodes(child_node):
                            if subcomponent.data_type == 'url.hostname':
                                assert isinstance(subcomponent.value, str)
                                preceding_domain = subcomponent.value
                                break
            else:
                preceding_domain = self.find_preceding_domain(parent_node)

        return preceding_domain

    def get_id(self):
        new_id = self.next_id
        self.next_id += 1
        return new_id

    def create_node(
            self, data_type, key, value, label, hover, parent_id=None,
            incoming_edge_config=None, extra_options=None):
        new_node = self.Node(
            self.get_id(), data_type=data_type, key=key, value=value,
            label=label, hover=hover, parent_id=parent_id,
            incoming_edge_config=incoming_edge_config,
            extra_options=extra_options)
        assert new_node.node_id not in self.nodes.keys()
        self.nodes[new_node.node_id] = new_node
        self.graph.add_node(new_node)
        self.total_nodes += 1

        if parent_id:
            if isinstance(parent_id, list):
                for parent in parent_id:
                    self.graph.add_edge(self.nodes[parent], new_node)
            else:
                self.graph.add_edge(self.nodes[parent_id], new_node)

        return new_node.node_id

    @staticmethod
    def add_b64_padding(encoded_string):
        encoded_string = encoded_string.rstrip('=')
        remainder = len(encoded_string) % 4
        if remainder == 1:
            return False
        elif remainder == 2:
            return f'{encoded_string}=='
        elif remainder == 3:
            return f'{encoded_string}='
        else:
            return encoded_string

    @staticmethod
    def check_if_int_between(value, low, high):
        try:
            value = int(value)
        except:
            return False

        if low < value < high:
            return True
        else:
            return False

    def add_to_queue(
            self, data_type, key, value, label=None, hover=None,
            parent_id=None, incoming_edge_config=None, extra_options=None):
        new_item = {
            'data_type': data_type,
            'key': key,
            'value': value,
            'label': label,
            'hover': hover,
            'incoming_edge_config': incoming_edge_config,
            'extra_options': extra_options
        }

        if parent_id:
            new_item['parent_id'] = parent_id

        if not extra_options:
            max_row_length = len(str(value)) * 2.2
            new_item['extra_options'] = \
                {'widthConstraint': {'maximum': max(max_row_length, 200)}}

        if new_item['data_type'] == 'url':
            new_item['value'] = re.sub(' ', '', new_item['value'])

        log.info(f'Added to queue: {new_item}')
        self.queue.put(new_item)

    def run_plugins(self, node):

        for unfurl_parser in unfurl.parsers.__all__:
            try:
                parser = importlib.import_module(f'unfurl.parsers.{unfurl_parser}')
            except ImportError as e:
                log.exception(f'Failed to import {unfurl_parser}: {e}')
                continue

            try:
                parser.run(self, node)
            except Exception as e:
                log.exception(f'Exception in {unfurl_parser}: {e}')

    def parse(self, queued_item):
        item = queued_item
        node_id = self.create_node(
            data_type=item['data_type'], key=item['key'], value=item['value'],
            label=item['label'], hover=utils.wrap_hover_text(item['hover']),
            parent_id=item.get('parent_id', None),
            incoming_edge_config=item.get('incoming_edge_config', None),
            extra_options=item.get('extra_options', None))

        self.run_plugins(self.nodes[node_id])

    def parse_queue(self):
        while not self.queue.empty() and self.total_nodes < 100:
            self.parse(self.queue.get())

    @staticmethod
    def transform_node(node):
        transformed = {
            'id': int(node.node_id),
            'label': f'{node.label}'
        }
        if node.hover:
            transformed['title'] = node.hover
        if node.extra_options:
            transformed.update(node.extra_options)
        return transformed

    @staticmethod
    def transform_edge(edge):
        transformed = {
            'from': int(edge[0].node_id),
            'to': int(edge[1].node_id)
        }

        if edge[1].incoming_edge_config:
            transformed.update(edge[1].incoming_edge_config)
        return transformed

    def generate_json(self):
        data_json = {'nodes': [], 'edges': []}
        for orig_node in self.graph.nodes():
            data_json['nodes'].append(self.transform_node(orig_node))
        for orig_edge in self.graph.edges():
            data_json['edges'].append(self.transform_edge(orig_edge))

        edge_summary = {}
        for edge in data_json.get('edges'):
            edge_summary.setdefault(edge.get('title'), 0)
            edge_summary[edge.get('title')] += 1

        data_json['summary'] = edge_summary

        return data_json

    @staticmethod
    def transform_3d_node(node):
        def val_func(node_id):
            if node_id == 1:
                return 15
            elif node_id < 10:
                return 10
            else:
                return 5

        def shorten_name(node_string):
            node_string = str(node_string)
            if len(node_string) > 60:
                return f'{node_string[:25]}...{node_string[-25:]}'
            else:
                return node_string

        node_color = '#aabfad'
        if node.incoming_edge_config:
            if node.incoming_edge_config['color']:
                node_color = node.incoming_edge_config['color']['color']

        transformed = {
            'id': str(node.node_id),
            'name': shorten_name(node.label),
            'fullName': f'{node.label}',
            'dataType': f'{node.data_type}',
            'val': val_func(node.node_id),
            'color': node_color
        }

        if node.hover:
            transformed['description'] = re.sub(r'<.*?>|\[.*?\]', '', node.hover)

        return transformed

    @staticmethod
    def transform_3d_edge(edge):
        transformed = {
            'source': str(edge[0].node_id),
            'target': str(edge[1].node_id)
        }

        if edge[1].incoming_edge_config:
            transformed.update(edge[1].incoming_edge_config)

        if transformed.get('color', {}).get('color'):
            transformed['color'] = transformed['color']['color']
        return transformed

    def generate_3d_json(self):
        data_json = {'nodes': [], 'links': []}
        for orig_node in self.graph.nodes():
            data_json['nodes'].append(self.transform_3d_node(orig_node))
        for orig_edge in self.graph.edges():
            data_json['links'].append(self.transform_3d_edge(orig_edge))
        return data_json

    def generate_text_tree(self, detailed=False, output_filter=None):
        tree_root = None
        for node_contents in self.graph.nodes(data=True):
            # Get the root node; id is 1. Needed for networkx tree_data().
            if node_contents[0].__dict__.get('node_id') == 1:
                tree_root = node_contents[0]
                break

        tree_data = networkx.readwrite.json_graph.tree_data(
            self.graph, root=tree_root)
        output_tree = Unfurl.text_tree(tree_data, detailed=detailed)

        if output_filter:
            filtered_tree = ''
            for line in output_tree.splitlines():
                if re.search(output_filter, line):
                    filtered_tree += f'\n{line}'
            output_tree = filtered_tree

        return output_tree

    @staticmethod
    def text_tree(tree_data, indent='', last_child=False, text_output='', detailed=False):

        node = tree_data['id']
        label = re.sub(r'\n', ' ', str(node.label))

        if node.node_id == 1:
            # This is the root node; don't indent to save space
            text_output += f'[{node.node_id}] {label}'
            if detailed:
                text_output += f' (type: {node.data_type})'
                if node.hover:
                    hover = re.sub(r'<.*?>|\[.*?\]', '', node.hover)
                    text_output += f' -- {hover}'

            indent += ' '

        elif not last_child:
            text_output += f'\n{indent}├─({node.incoming_edge_config["label"]})─[{node.node_id}] {label}'
            if detailed:
                text_output += f' (type: {node.data_type})'
                if node.hover:
                    hover = re.sub(r'<.*?>|\[.*?\]', '', node.hover)
                    text_output += f' -- {hover}'

            indent += '|  '
        else:
            text_output += f'\n{indent}└─({node.incoming_edge_config["label"]})─[{node.node_id}] {label}'
            if detailed:
                text_output += f' (type: {node.data_type})'
                if node.hover:
                    hover = re.sub(r'<.*?>|\[.*?\]', '', node.hover)
                    text_output += f' -- {hover}'

            indent += '   '

        if tree_data.get('children'):
            children_text_output = ''
            for number, child in enumerate(tree_data['children']):
                last_child = \
                    True if (number + 1) == len(tree_data['children']) else False
                children_text_output += \
                    Unfurl.text_tree(
                        child, indent=indent, last_child=last_child, detailed=detailed)
            text_output += children_text_output

        return text_output


unfurl_app_host = None
unfurl_app_port = None
unfurl_remote_lookups = None
app = Flask(__name__)
CORS(app)


class UnfurlApp:
    def __init__(self, unfurl_debug='True', unfurl_host='localhost', unfurl_port='5000', remote_lookups=False):
        self.unfurl_debug = unfurl_debug
        self.unfurl_host = unfurl_host
        self.unfurl_port = unfurl_port
        self.remote_lookups = remote_lookups

        global unfurl_app_host
        global unfurl_app_port
        global unfurl_remote_lookups
        unfurl_app_host = self.unfurl_host
        unfurl_app_port = self.unfurl_port
        unfurl_remote_lookups = self.remote_lookups

        app.config['remote_lookups'] = remote_lookups
        app.run(debug=unfurl_debug, host=unfurl_host, port=unfurl_port)


@app.route('/')
def index():
    return render_template(
        'graph.html', url_to_unfurl='', unfurl_host=unfurl_app_host,
        unfurl_port=unfurl_app_port)


@app.route('/<path:url_to_unfurl>')
def graph(url_to_unfurl):
    return render_template(
        'graph.html', url_to_unfurl=url_to_unfurl,
        unfurl_host=unfurl_app_host, unfurl_port=unfurl_app_port)


@app.route('/api/<path:api_path>')
def api(api_path):
    # Get the referrer from the request, which has the full url + query string.
    # Split off the local server and keep just the url we want to parse
    unfurl_this = request.referrer.split(f':{unfurl_app_port}/', 1)[1]

    unfurl_instance = Unfurl(remote_lookups=app.config['remote_lookups'])
    unfurl_instance.add_to_queue(
        data_type='url', key=None,
        extra_options={'widthConstraint': {'maximum': 1200}},
        value=unfurl_this)
    unfurl_instance.parse_queue()

    unfurl_json = unfurl_instance.generate_json()
    return unfurl_json



#!/usr/bin/env python3

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import csv
import os
from unfurl import core


def main():
    parser = argparse.ArgumentParser(
        description='unfurl takes a URL and expands ("unfurls") it into a directed graph, extracting every '
                    'bit of information from the URL and exposing the obscured.')
    parser.add_argument(
        'what_to_unfurl',
        help='what to unfurl. typically this is a URL, but it also supports integers (timestamps), '
             'encoded protobufs, and more. if this is instead a file path, unfurl will open '
             'that file and process each line in it as a URL.')
    parser.add_argument(
        '-d', '--detailed', help='show more detailed explanations.', action='store_true')
    parser.add_argument(
        '-f', '--filter', help='only output lines that match this filter.')
    parser.add_argument(
        '-l', '--lookups', help='allow remote lookups to enhance results.', action='store_true')
    parser.add_argument(
        '-o', '--output',
        help='file to save output (as CSV) to. if omitted, output is sent to '
             'stdout (typically this means displayed in the console).')
    parser.add_argument(
        '-v', '-V', '--version', action='version', version=f'unfurl v{core.unfurl.__version__}')
    args = parser.parse_args()

    items_to_unfurl = []

    if os.path.isfile(args.what_to_unfurl):
        with open(args.what_to_unfurl, errors='ignore') as f:
            for input_url in f:
                items_to_unfurl.append(input_url.rstrip())

    else:
        items_to_unfurl.append(args.what_to_unfurl)

    if args.output:
        with open(args.output, 'w', newline='', encoding='utf-8') as csv_file:
            csv_writer = csv.writer(csv_file, quoting=csv.QUOTE_ALL)
            csv_writer.writerow(['url', 'unfurled'])

            for item in items_to_unfurl:
                unfurl_instance = core.Unfurl(remote_lookups=args.lookups)
                unfurl_instance.add_to_queue(
                    data_type='url', key=None,
                    value=item)
                unfurl_instance.parse_queue()
                csv_writer.writerow(
                    [item, unfurl_instance.generate_text_tree(
                        detailed=args.detailed,
                        output_filter=args.filter)])

    else:
        for item in items_to_unfurl:
            unfurl_instance = core.Unfurl(remote_lookups=args.lookups)
            unfurl_instance.add_to_queue(
                data_type='url', key=None,
                value=item)
            unfurl_instance.parse_queue()

            print(unfurl_instance.generate_text_tree(
                detailed=args.detailed, output_filter=args.filter))
            print()


if __name__ == "__main__":
    main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import uuid

uuid_edge = {
    'color': {
        'color': 'orange'
    },
    'title': 'UUID Parsing Functions',
    'label': 'UU'
}


def run(unfurl, node):
    if not node.data_type.startswith(('uuid', 'hash')):
        # Leading '/' optional, 8-4-4-4-12 hex digits with '-' optional, 13th char limited to [1-5]
        m = re.fullmatch(r'/?([0-9A-F]{8}-?[0-9A-F]{4}-?[1-5][0-9A-F]{3}-?[0-9A-F]{4}-?[0-9A-F]{12})',
                         str(node.value), re.IGNORECASE)
        if m:
            u = m.group(1)
            u = u.replace('-', '')
            unfurl.add_to_queue(
                data_type='uuid', key=None, value=u, label=f'UUID: {u[:8]}-{u[8:12]}-{u[12:16]}-{u[16:20]}-{u[20:]}',
                parent_id=node.node_id, incoming_edge_config=uuid_edge,
                hover='This is potentially a UUID, based on length',
                extra_options={'widthConstraint': {'maximum': 500}})

    elif node.data_type == 'uuid':
        u = uuid.UUID(node.value)
        # TODO: reference: http://www.mahonri.info/cgi/uuid.cgi
        if u.version == 1:
            unfurl.add_to_queue(
                data_type='uuid-parsed', key=None, value=node.value,
                label='Version 1 UUID is based on time and MAC address',
                parent_id=node.node_id, incoming_edge_config=uuid_edge)

            # Timestamp is 60-bits, the number of 100-nanosecond intervals since midnight 15 October 1582 Coordinated
            # Universal Time (UTC). The offset to the Unix epoch is 12219292800000 milliseconds. Divide ts by 10000
            # to go from 100-ns intervals to milliseconds.x
            timestamp = u.time/10000 - 12219292800000
            unfurl.add_to_queue(
                data_type='epoch-milliseconds', key=None, value=timestamp, label=f'Time generated: {timestamp}',
                parent_id=node.node_id, incoming_edge_config=uuid_edge)

            # Detection for randomly generated MACs (random 48-bit number with its eighth bit set to 1 as
            #  recommended in RFC 4122)
            if (u.node >> 40) & 0x01 == 1:
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value='The Node ID in this UUID is random',
                    hover='The Node ID in this UUID is set to a random number, <br>rather than an actual MAC address. '
                          '<a href="https://tools.ietf.org/html/rfc4122#section-4.5">[ref]</a>',
                    parent_id=node.node_id, incoming_edge_config=uuid_edge)

            else:
                pretty_mac = f'{u.node:0{12}X}'
                pretty_mac = ':'.join([pretty_mac[i]+pretty_mac[i+1] for i in range(0, 12, 2)])

                unfurl.add_to_queue(
                    data_type='mac-address', key=None, value=pretty_mac, label=f'MAC address: {pretty_mac}',
                    parent_id=node.node_id, incoming_edge_config=uuid_edge)

        elif u.version == 3:
            unfurl.add_to_queue(
                data_type='uuid-parsed', key=None, value=node.value,
                label='Version 3 UUID is generated based on a namespace and a name, which are '
                      'combined and hashed using MD5', parent_id=node.node_id, incoming_edge_config=uuid_edge)

        elif u.version == 4:
            # Limits to timestamp values between 2018-01 and 2025-05
            if 0x8A <= int(node.value[:2], 16) <= 0x9F:
                # Ref: https://itnext.io/laravel-the-mysterious-ordered-uuid-29e7500b4f8
                unfurl.add_to_queue(
                    data_type='descriptor', key=None,
                    value='This UUID matches the structure of both UUIDv4 (random) '
                          '& Laravel "Ordered UUIDs".',
                    hover='Laravel Ordered UUIDs appear similar to UUIDv4s, but they are outside RFC 4122 '
                          '<a href="https://itnext.io/laravel-the-mysterious-ordered-uuid-29e7500b4f8">[ref]</a>.'
                          '<br>The "Ordered UUID" is composed of a timestamp and random bits, while UUIDv4 is '
                          '<br>random and does not contain a timestamp.'
                          '<br><br>Unfurl differentiates between them based on potential timestamp values. There is '
                          '<br>a chance of misidentifying them, as they use the same structure. UUIDv4 is more common, '
                          '<br>but Unfurl is parsing this as an "Ordered UUID" to show the potential timestamp.',
                    extra_options= {'widthConstraint': {'maximum': 250}},
                    parent_id=node.node_id, incoming_edge_config=uuid_edge)

                unfurl.add_to_queue(
                    data_type='epoch-ten-microseconds', key='Timestamp', value=int(node.value[:12], 16),
                    hover='The first 48 bits of an Ordered UUID are a timestamp',
                    parent_id=node.node_id, incoming_edge_config=uuid_edge)

            else:
                unfurl.add_to_queue(
                    data_type='uuid-parsed', key=None, value=node.value, label='Version 4 UUID is randomly generated',
                    parent_id=node.node_id, incoming_edge_config=uuid_edge)

        elif u.hex[12] == '5':
            unfurl.add_to_queue(
                data_type='uuid-parsed', key=None, value=node.value,
                label='Version 5 UUID is generated based on a namespace and a name, which are '
                      'combined and hashed using SHA-1', parent_id=node.node_id, incoming_edge_config=uuid_edge)

    # TODO: Add lookup of well-known GUIDs?
    # https://github.com/EricZimmerman/GuidMapping/blob/master/Resources/GuidToName.txt



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging
log = logging.getLogger(__name__)

twitter_snowflake_edge = {
    'color': {
        'color': '#1da1f2'
    },
    'title': 'Twitter Snowflake',
    'label': '❄'
}

twitter_edge = {
    'color': {
        'color': '#1da1f2'
    },
    'title': 'Twitter Parsing Functions',
    'label': '@'
}


def parse_twitter_snowflake(unfurl, node, encoding_type='integer', on_twitter=True):
    # Ref: https://blog.twitter.com/engineering/en_us/a/2010/announcing-snowflake.html
    #      https://ws-dl.blogspot.com/2019/08/2019-08-03-tweetedat-finding-tweet.html
    if encoding_type == 'integer':
        try:
            snowflake = int(node.value)
            timestamp = (snowflake >> 22) + 1288834974657
            machine_id = (snowflake & 0x3FF000) >> 12
            sequence = snowflake & 0xFFF
            assert 0 <= sequence <= 4096, \
                f'Sequence number should be between 0 and 4096; got {sequence}'

        except Exception as e:
            log.exception(f'Exception parsing snowflake: {e}')
            return

    # Ref: https://www.hackerfactor.com/blog/index.php?/archives/634-Name-Dropping.html
    # So far, I've only observed this in file attachments (images) to tweets. This format
    # also has three extra bytes at the "end" (LSBs); I don't know their function.
    elif encoding_type == 'base64':
        try:
            padded_value = unfurl.add_b64_padding(node.value)
            snowflake_bytes = base64.urlsafe_b64decode(padded_value)
            snowflake_int = int.from_bytes(snowflake_bytes, 'big')
            timestamp = (snowflake_int >> 46) + 1288834974657
            machine_id = (snowflake_int & 0x3FF000000000) >> 36
            sequence = (snowflake_int & 0xFFF000000) >> 24
            assert 0 <= sequence <= 4096, \
                f'Sequence number should be between 0 and 4096; got {sequence}'

            # Since we are trying to parse things that might not be valid, make sure the decoded
            # timestamp is "reasonable" (between 2010-11 (the Snowflake epoch) and 2024-01
            if not 1288834974657 < timestamp < 1704070800000:
                return

        except Exception as e:
            log.warning(f'Exception parsing snowflake: {e}')
            return
    else:
        return

    if on_twitter:
        node.hover = 'Twitter Snowflakes are time-based IDs. ' \
                     '<a href="https://blog.twitter.com/engineering/en_us/a/2010/announcing-snowflake.html" ' \
                     'target="_blank">[ref]</a>'

    else:
        unfurl.add_to_queue(
            data_type='descriptor', key=None, value='File may originally be from Twitter (based on file name)',
            hover='File name is consistent with Twitter\'s naming scheme for image attachments. The name also contains '
                  'an embedded timestamp.', parent_id=node.node_id, incoming_edge_config=twitter_snowflake_edge)

    unfurl.add_to_queue(
        data_type='epoch-milliseconds', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
        hover='The first value in a Twitter Snowflake is a timestamp',
        parent_id=node.node_id, incoming_edge_config=twitter_snowflake_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=machine_id, label=f'Machine ID: {machine_id}',
        hover='The second value in a Twitter Snowflake is the machine ID. A machine ID can be further divided '
              'into a datacenter ID (first 5 bits) and worker ID (last 5 bits)',
        parent_id=node.node_id, incoming_edge_config=twitter_snowflake_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=sequence, label=f'Sequence: {sequence}',
        hover='For every ID that is generated, this number is incremented and rolls over every 4096',
        parent_id=node.node_id, incoming_edge_config=twitter_snowflake_edge)


def run(unfurl, node):
    preceding_domain = unfurl.find_preceding_domain(node)
    if preceding_domain == 'twitter.com':
        # Make sure potential snowflake is reasonable: between 2015-02-01 & 2024-06-10
        if node.data_type == 'url.path.segment' and \
                unfurl.check_if_int_between(node.value, 261675293291446272, 1800000000000000001):
            parse_twitter_snowflake(unfurl, node)

        elif node.data_type == 'url.query.pair':
            if node.key == 's':
                sharing_codes = {
                    '19': ' from Android',
                    '20': ' from Twitter Web',
                    '21': ' from iOS'
                }

                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Tweet was shared{sharing_codes.get(node.value, "")}',
                    hover='The sharing platform is a best guess based on observations; if you find a '
                          'new sharing code, please let us know',
                    parent_id=node.node_id, incoming_edge_config=twitter_edge)

    # I've only seen `pbs.twimg.com`, but there could be other subdomains used for image attachments to tweets
    elif '.twimg.com' in preceding_domain:
        if node.data_type == 'url.path.segment':
            # Viewing an image attached to a tweet
            # Ex: https://pbs.twimg.com/media/ErD8xv0VkAIBJe7?format=jpg&name=medium
            if unfurl.check_sibling_nodes(node, data_type='url.path.segment', key=1, value='media'):
                if node.key == 2:
                    parse_twitter_snowflake(unfurl, node, encoding_type='base64')

    # It's possible an image from Twitter was saved off and then uploaded somewhere else. The file name
    # pattern appears fairly unique, so if we see a file name that matches it and decodes to a "reasonable"
    # timestamp, show it in the graph.
    elif node.data_type == 'file.name' and len(node.value) == 15:
        parse_twitter_snowflake(unfurl, node, encoding_type='base64', on_twitter=False)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


dropbox_edge = {
    'color': {
        'color': '#2C6AFB'
    },
    'title': 'Dropbox Parsing Functions',
    'label': 'DB'
}


def run(unfurl, node):
    # References: https://www.atropos4n6.com/cloud-forensics/artifacts-of-dropbox-usage-on-windows-10-part-2/
    if 'dropbox.com' in unfurl.find_preceding_domain(node):
        if node.data_type == 'url.path' and node.value.startswith('/home'):
            _, viewed_directory = node.value.split('/home', 1)
            viewed_value = ''
            if viewed_directory and viewed_directory != '/':
                viewed_value = f'directory "{viewed_directory[1:]}" from '
            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=f'Viewing {viewed_value}the user\'s "All Files" page',
                parent_id=node.node_id, incoming_edge_config=dropbox_edge)

        elif node.data_type == 'url.path' and node.value.startswith('/h'):
            _, viewed_directory = node.value.split('/h', 1)
            viewed_value = ''
            if viewed_directory and viewed_directory != '/':
                viewed_value = f'directory "{viewed_directory[1:]}" from '
            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=f'Viewing {viewed_value}the user\'s Dropbox "Home" page',
                hover='The Dropbox "Home" page has suggestions, recent, and starred items.',
                parent_id=node.node_id, incoming_edge_config=dropbox_edge)

        elif node.data_type == 'url.query.pair':
            if node.key == 'preview':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'File "{node.value}" is being previewed',
                    parent_id=node.node_id, incoming_edge_config=dropbox_edge)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torf

magnet_edge = {
    'color': {
        'color': '#CF0024'
    },
    'title': 'Magnet-related Parsing Functions',
    'label': '🧲'
}

known_field_names = {
    'as_': {'name': 'Acceptable Source',
            'description': '<b>as_ (Acceptable Source)</b> is the web link to the file online'},
    'dn': {'name': 'Display Name',
           'description': '<b>dn (Display Name)</b> is the file name to display to the user, for convenience'},
    'kt': {'name': 'Keyword Topic',
           'description': '<b>kt (Keyword Topic)</b> is a more general search,'
                          ' specifying keywords, rather than a particular file'},
    'tr': {'name': 'Address Tracker',
           'description': '<b>tr (Address Tracker)</b> is the tracker URL for BitTorrent downloads'},
    'ws': {'name': 'Web Seed', 'description': '<b>ws (Web Seed)</b> is the payload data served over HTTP(S)'},
    'xl': {'name': 'Exact Length', 'description': '<b>xl (Exact Length)</b> is the size in bytes'},
    'xs': {'name': 'Exact Source',
           'description': '<b>xs (Exact Source)</b> is the P2P link identified by a content-hash'},
    'xt': {'name': 'Exact Topic', 'description': '<b>xt (Exact Topic)</b> is the URN containing file hash'}
}

# Based off https://en.wikipedia.org/wiki/Magnet_URI_scheme#BitTorrent_info_hash_(BTIH)
xt_hash_types = {
    'tree': '<b>Tiger Tree Hash (TTH)</B>. '
            '<br>These hashes are used on Direct Connect and G2 (Gnutella2), among others.',
    'sha1': '<b>Secure Hash Algorithm 1 (SHA-1)</b>. <br>These hash sums are used on gnutella and G2 (Gnutella2).',
    'bitprint': '<b>BitPrint</b>. <br>Such hash sums consist of an SHA-1 Hash, followed by a TTH Hash, '
                'delimited by a point; they are used on gnutella and G2 (Gnutella2).',
    'ed2k': '<b>ED2K (eDonkey2000) hash</b>. <br>These hash sums are used on eDonkey2000.',
    'aich': '<b>Advanced Intelligent Corruption Handler (AICH)</b>. <br>Not formal URNs for Magnet links, such hashes '
            'are used by eDonkey2000 to restore and control the integrity of downloading and already downloaded files.',
    'kzhash': '<b>Kazaa hash</b>. <br>Used on FastTrack, these hash sums are vulnerable to hash collision attacks.',
    'btih': '<b>BitTorrent info hash (BTIH)</b>. <br>These are hex-encoded SHA-1 hash sums of the "info" sections '
            '<br>of BitTorrent metafiles as used by BitTorrent to identify downloadable <br>files or sets of files. '
            'For backwards compatibility with existing <br>links, clients should also support the Base32 encoded '
            'version of the hash.',
    'md5': '<b>Message Digest 5 (MD5)</b>. '
           '<br>Supported by G2 (Gnutella2), such hashes are vulnerable to hash collision attacks.',
}


def run(unfurl, node):
    if node.data_type.startswith('magnet') and node.data_type.endswith('list'):
        for node_item in node.value:
            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node_item,
                parent_id=node.node_id, incoming_edge_config=magnet_edge)

    elif node.data_type == 'magnet.xt':
        urn, xt_type, xt_hash = node.value.split(':', 2)
        if xt_type in xt_hash_types.keys():
            unfurl.add_to_queue(
                data_type='descriptor', key='xt Hash Type', value=xt_type,
                hover=xt_hash_types[xt_type], parent_id=node.node_id,
                incoming_edge_config=magnet_edge)

        unfurl.add_to_queue(
            data_type='xt.hash', key='xt.hash', value=xt_hash,
            hover='This <b>xt hash</b> is used to find and verify the files specified in the Magnet link.',
            parent_id=node.node_id, incoming_edge_config=magnet_edge)

    elif isinstance(node.value, str):
        if node.value.startswith('magnet:?'):

            # TODO: torf won't parse xt.1, xt.2, etc fields, but those are allowed
            # ref: https://en.wikipedia.org/wiki/Magnet_URI_scheme#BitTorrent_info_hash_(BTIH)
            parsed_magnet = torf.Magnet.from_string(node.value)

            for known_field in known_field_names.keys():
                if getattr(parsed_magnet, known_field):
                    field_value = getattr(parsed_magnet, known_field)

                    if isinstance(field_value, torf._utils.MonitoredList):
                        if len(field_value) == 1:
                            unfurl.add_to_queue(
                                data_type=f'magnet.{known_field}', key=known_field_names[known_field]['name'],
                                value=field_value[0],
                                hover=known_field_names[known_field]['description'],
                                parent_id=node.node_id, incoming_edge_config=magnet_edge)
                        elif len(field_value) > 1:
                            unfurl.add_to_queue(
                                data_type=f'magnet.{known_field}.list', key=known_field_names[known_field]['name'],
                                value=list(field_value),
                                hover=known_field_names[known_field]['description'],
                                parent_id=node.node_id, incoming_edge_config=magnet_edge)

                    elif field_value:
                        unfurl.add_to_queue(
                            data_type=f'magnet.{known_field}', key=known_field_names[known_field]['name'],
                            value=field_value,
                            hover=known_field_names[known_field]['description'],
                            parent_id=node.node_id, incoming_edge_config=magnet_edge)

            # TODO: experimental x.* fields
            for field, value in parsed_magnet.__dict__.items():
                if field in [f'_{known}' for known in known_field_names.keys()]:
                    continue
                # 'infohash' is extracted from xt by torf and added as it's own thing (for convenience); skip it
                elif field == '_infohash':
                    continue
                elif value:
                    unfurl.add_to_queue(
                        data_type='descriptor', key=field, value=value,
                        parent_id=node.node_id, incoming_edge_config=magnet_edge)



#!/usr/bin/env python3

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import textwrap
from typing import Union

long_int_re = re.compile(r'\d{8,}')
urlsafe_b64_re = re.compile(r'[A-Za-z0-9_\-]{8,}={0,2}')
standard_b64_re = re.compile(r'[A-Za-z0-9+/]{8,}={0,2}')
hex_re = re.compile(r'([A-F0-9]{2})+', flags=re.IGNORECASE)
digits_re = re.compile(r'\d+')
letters_re = re.compile(r'[A-Z]+', flags=re.IGNORECASE)
float_re = re.compile(r'\d+\.\d+')
mac_addr_re = re.compile(r'(?P<mac_addr>[0-9A-Fa-f]{12}|([0-9A-Fa-f]:){6})')
cisco_7_re = re.compile(r'\d{2}[A-F0-9]{4,}', re.IGNORECASE)


def wrap_hover_text(hover_text: Union[str, None]) -> Union[str, None]:
    if not hover_text:
        return None

    if not isinstance(hover_text, str):
        return None

    # If there are any manually-inserted <br> or links, leave it
    # alone. This isn't perfect detection, but it'll do.
    if '<a' in hover_text or '<br' in hover_text:
        return hover_text

    # If the text is just a little long, I'd rather have it all on
    # one line than split into one long line and a short second line
    if len(hover_text) < 70:
        return hover_text

    # "Wrap" the hover text by splitting it into lines of length <width>,
    # then joining them together with a <br>.
    return '<br>'.join(textwrap.wrap(hover_text, width=60))



# Copyright 2020 Moshe Kaplan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License athttps://channel9.msdn.com/Shows/Going+Deep?page=20
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

yahoo = {
    'color': {
        'color': '#6001d2'
    },
    'title': 'Yahoo Search-related Parsing Functions',
    'label': 'Y!'
}


def run(unfurl, node):
    if node.data_type == 'url.query.pair':
        if 'yahoo.com' in unfurl.find_preceding_domain(node):

            if node.key == 'durs':
                durs_mappings = {
                    'short': 'Short (less than 5 minutes)',
                    'medium': 'Medium (5-20 minutes)',
                    'long': 'Long (more than 20 minutes)',
                }
                value = durs_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Video Duration: {value}',
                    hover='Yahoo Video Search Duration', parent_id=node.node_id, incoming_edge_config=yahoo)

            elif node.key == 'imgc':
                imgc_mappings = {
                    'black': 'Black',
                    'blue': 'Blue',
                    'brown': 'Brown',
                    'bw': 'Black & White',
                    'gray': 'Gray',
                    'green': 'Green',
                    'orange': 'Orange',
                    'pink': 'Pink',
                    'purple': 'Purple',
                    'red': 'Red',
                    'teal': 'Teal',
                    'white': 'White',
                    'yellow': 'Yellow',

                }
                value = imgc_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Image Size: {value}',
                    hover='Yahoo Image Search Size', parent_id=node.node_id, incoming_edge_config=yahoo)                    

            elif node.key == 'imgsz':
                imgsz_mappings = {
                    'small': 'Small',
                    'medium': 'Medium',
                    'large': 'Large',
                }
                value = imgsz_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Image Size: {value}',
                    hover='Yahoo Image Search Size', parent_id=node.node_id, incoming_edge_config=yahoo)
                    
            elif node.key == 'imgty':
                imgty_mappings = {
                    'clipart': 'Clipart',
                    'face': 'Face',
                    'gif': 'GIF',
                    'graphics': 'Graphics',
                    'linedrawing': 'Line Drawing',
                    'nonportrait': 'Non Portrait',
                    'photo': 'Photo',
                    'portrait': 'Portrait',
                }
                value = imgty_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Image Type: {value}',
                    hover='Yahoo Image Search Type', parent_id=node.node_id, incoming_edge_config=yahoo)       
                    
            elif node.key == 'p':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Search Query: {node.value}',
                    hover='Terms used in the Yahoo search', parent_id=node.node_id, incoming_edge_config=yahoo)

            elif node.key == 'vage':
                vage_mappings = {
                    'day': 'Past 24 hours',
                    'month': 'Past month',
                    'year': 'Past year',
                }
                value = vage_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Video Duration: {value}',
                    hover='Yahoo Video Search Duration', parent_id=node.node_id, incoming_edge_config=yahoo)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
log = logging.getLogger(__name__)

tiktok_edge = {
    'color': {
        'color': '#F60051'
    },
    'title': 'TikTok Parsing Functions',
    'label': '♪'
}


def parse_tiktok_id(unfurl, node):
    try:
        tiktok_id = int(node.value)
        # TikTok IDs are 63 bits long; the upper 31 bits are the timestamp.
        # Shift it 32 (63-32 = 31) bits right, leaving just the timestamp bits.
        timestamp = (tiktok_id >> 32)

    except Exception as e:
        log.exception(f'Exception parsing TikTok ID: {e}')
        return

    node.hover = 'TikTok IDs are time-based IDs similar to those of Twitter Snowflakes.'

    unfurl.add_to_queue(
        data_type='epoch-seconds', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
        hover='The leading 31 bits in a TikTok ID are a timestamp, '
              '<br>thought to represent when an item was posted or created.',
        parent_id=node.node_id, incoming_edge_config=tiktok_edge)


def run(unfurl, node):

    if node.data_type == 'url.path.segment':

        if 'tiktok.com' in unfurl.find_preceding_domain(node):
            if node.key == 1:
                if node.value.startswith('@'):
                    unfurl.add_to_queue(
                        data_type='descriptor', key=None, value=f'Username who posted on TikTok',
                        parent_id=node.node_id, incoming_edge_config=tiktok_edge)
                elif node.value == 'embed':
                    unfurl.add_to_queue(
                        data_type='descriptor', key=None, value=f'TikTok video was embedded on another site',
                        parent_id=node.node_id, incoming_edge_config=tiktok_edge)

            # Check if TikTok ID timestamp would be between 2017-12 and 2025-05
            elif unfurl.check_if_int_between(node.value, 6500000000000000000, 7500000000000000000):
                parse_tiktok_id(unfurl, node)

    # If it's the "root" node and a plausible TikTok ID, parse it.
    # This case covers someone parsing just a TikTok ID, not a full URL.
    elif node.node_id == 1 and unfurl.check_if_int_between(node.value, 6500000000000000000, 7500000000000000000):
        parse_tiktok_id(unfurl, node)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import maclookup
import re
from unfurl import utils

uuid_edge = {
    'color': {
        'color': 'red'
    },
    'title': 'MAC Parsing Functions',
    'label': '🖧'
}


def run(unfurl, node):
    if not node.data_type == 'mac-address':
        long_int = utils.long_int_re.fullmatch(str(node.value))
        m = utils.mac_addr_re.fullmatch(str(node.value))
        if m and not long_int:
            u = m.group('mac_addr')

            # Check if we need to add colons
            if len(u) == 12:
                pretty_mac = ':'.join([u[i]+u[i+1] for i in range(0, 12, 2)])

            else:
                pretty_mac = u.upper()

            unfurl.add_to_queue(
                data_type='mac-address', key=None, value=pretty_mac, label=f'MAC address: {pretty_mac}',
                parent_id=node.node_id, incoming_edge_config=uuid_edge)

    elif node.data_type == 'mac-address' and unfurl.api_keys.get('macaddress_io') and unfurl.remote_lookups:
        client = maclookup.ApiClient(unfurl.api_keys.get('macaddress_io'))
        vendor_lookup = client.get_vendor(node.value).decode('utf-8')

        if vendor_lookup:
            unfurl.add_to_queue(
                data_type="mac-address.vendor", key=None, value=vendor_lookup, label=f'Vendor: {vendor_lookup}',
                parent_id=node.node_id, incoming_edge_config=uuid_edge)



"""Contains various maps for protobuf types, including encoding/decoding
   functions, wiretypes and default types
"""
from google.protobuf.internal import wire_format
from unfurl.lib.blackboxprotobuf.lib.types import fixed, length_delim
from unfurl.lib.blackboxprotobuf.lib.types import varint

ENCODERS = {
    "uint": varint.encode_uvarint,
    "int": varint.encode_varint,
    "sint": varint.encode_svarint,
    "fixed32": fixed.encode_fixed32,
    "sfixed32": fixed.encode_sfixed32,
    "float": fixed.encode_float,
    "fixed64": fixed.encode_fixed64,
    "sfixed64": fixed.encode_sfixed64,
    "double": fixed.encode_double,
    "bytes": length_delim.encode_bytes,
    "bytes_hex": length_delim.encode_bytes_hex,
    "string": length_delim.encode_string,
    "packed_uint": length_delim.generate_packed_encoder(varint.encode_uvarint),
    "packed_int": length_delim.generate_packed_encoder(varint.encode_varint),
    "packed_sint": length_delim.generate_packed_encoder(varint.encode_svarint),
    "packed_fixed32": length_delim.generate_packed_encoder(fixed.encode_fixed32),
    "packed_sfixed32": length_delim.generate_packed_encoder(fixed.encode_sfixed32),
    "packed_float": length_delim.generate_packed_encoder(fixed.encode_float),
    "packed_fixed64": length_delim.generate_packed_encoder(fixed.encode_fixed64),
    "packed_sfixed64": length_delim.generate_packed_encoder(fixed.encode_sfixed64),
    "packed_double": length_delim.generate_packed_encoder(fixed.encode_double),
}

DECODERS = {
    "uint": varint.decode_uvarint,
    "int": varint.decode_varint,
    "sint": varint.decode_svarint,
    "fixed32": fixed.decode_fixed32,
    "sfixed32": fixed.decode_sfixed32,
    "float": fixed.decode_float,
    "fixed64": fixed.decode_fixed64,
    "sfixed64": fixed.decode_sfixed64,
    "double": fixed.decode_double,
    "bytes": length_delim.decode_bytes,
    "bytes_hex": length_delim.decode_bytes_hex,
    "string": length_delim.decode_string,
    "packed_uint": length_delim.generate_packed_decoder(varint.decode_uvarint),
    "packed_int": length_delim.generate_packed_decoder(varint.decode_varint),
    "packed_sint": length_delim.generate_packed_decoder(varint.decode_svarint),
    "packed_fixed32": length_delim.generate_packed_decoder(fixed.decode_fixed32),
    "packed_sfixed32": length_delim.generate_packed_decoder(fixed.decode_sfixed32),
    "packed_float": length_delim.generate_packed_decoder(fixed.decode_float),
    "packed_fixed64": length_delim.generate_packed_decoder(fixed.decode_fixed64),
    "packed_sfixed64": length_delim.generate_packed_decoder(fixed.decode_sfixed64),
    "packed_double": length_delim.generate_packed_decoder(fixed.decode_double),
}

WIRETYPES = {
    "uint": wire_format.WIRETYPE_VARINT,
    "int": wire_format.WIRETYPE_VARINT,
    "sint": wire_format.WIRETYPE_VARINT,
    "fixed32": wire_format.WIRETYPE_FIXED32,
    "sfixed32": wire_format.WIRETYPE_FIXED32,
    "float": wire_format.WIRETYPE_FIXED32,
    "fixed64": wire_format.WIRETYPE_FIXED64,
    "sfixed64": wire_format.WIRETYPE_FIXED64,
    "double": wire_format.WIRETYPE_FIXED64,
    "bytes": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "bytes_hex": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "string": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "message": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "group": wire_format.WIRETYPE_START_GROUP,
    "packed_uint": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_int": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_sint": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_fixed32": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_sfixed32": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_float": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_fixed64": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_sfixed64": wire_format.WIRETYPE_LENGTH_DELIMITED,
    "packed_double": wire_format.WIRETYPE_LENGTH_DELIMITED,
}

WIRE_TYPE_DEFAULTS = {
    wire_format.WIRETYPE_VARINT: "int",
    wire_format.WIRETYPE_FIXED32: "fixed32",
    wire_format.WIRETYPE_FIXED64: "fixed64",
    wire_format.WIRETYPE_LENGTH_DELIMITED: None,
    wire_format.WIRETYPE_START_GROUP: None,
    wire_format.WIRETYPE_END_GROUP: None,
}



"""Exception classes for BlackboxProtobuf"""


class BlackboxProtobufException(Exception):
    """Base class for excepions raised by Blackbox Protobuf"""

    def __init__(self, message, path=None, *args):
        self.path = path
        super(BlackboxProtobufException, self).__init__(message, *args)

    def set_path(self, path):
        if self.path is None:
            self.path = path


class TypedefException(BlackboxProtobufException):
    """Thrown when an error is identified in the type definition, such as
    conflicting or incosistent values."""

    def __str__(self):
        message = super(TypedefException, self).__str__()
        if self.path is not None:
            message = (
                "Encountered error within typdef for field %s: "
                % "->".join(map(str, self.path))
            ) + message
        return message


class EncoderException(BlackboxProtobufException, ValueError):
    """Thrown when there is an error encoding a dictionary to a type definition"""

    def __str__(self):
        message = super(EncoderException, self).__str__()
        if self.path is not None:
            message = (
                "Encountered error encoding field %s: " % "->".join(map(str, self.path))
            ) + message
        return message


class DecoderException(BlackboxProtobufException, ValueError):
    """Thrown when there is an error decoding a bytestring to a dictionary"""

    def __str__(self):
        message = super(DecoderException, self).__str__()
        if self.path is not None:
            message = (
                "Encountered error decoding field %s: " % "->".join(map(str, self.path))
            ) + message
        return message



# Copyright 2020 Moshe Kaplan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

duckduckgo = {
    'color': {
        'color': '#D74F38'
    },
    'title': 'DuckDuckGo-related Parsing Functions',
    'label': '🦆'
}


def run(unfurl, node):
    # Some URL parameters are from https://duckduckgo.com/params
    if node.data_type == 'url.query.pair':
        if 'duckduckgo' in unfurl.find_preceding_domain(node):

            if node.key == 'df':
                df_mappings = {
                    'd': 'Past Day',
                    'm': 'Past Month',
                    'w': 'Past Week',
                    'y': 'Past Year'
                }
                value = df_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Time Period: {value}',
                    hover='DuckDuckGo Search Period', parent_id=node.node_id, incoming_edge_config=duckduckgo)

            elif node.key == 'ia':
                ia_mappings = {
                    'images': 'Image Search',
                    'news': 'News Search',
                    'videos': 'Video Search',
                    'web': 'Web Search',
                    'definition': 'Definition',
                    'shopping': 'Shopping Search'
                }
                value = ia_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Search Type: {value}',
                    hover='DuckDuckGo Search Type', parent_id=node.node_id, incoming_edge_config=duckduckgo)

            elif node.key == 'q':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Search Query: {node.value}',
                    hover='Terms used in the DuckDuckGo search', parent_id=node.node_id, incoming_edge_config=duckduckgo)

            # ref: https://help.duckduckgo.com/privacy/t/
            elif node.key == 't':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Tracking Code: {node.value}',
                    hover='To assign advertising revenue and collect anonymous aggregate usage <br>information, '
                          'developers add a unique "&t=" parameter to searches <br>made through their applications. '
                          '<a href="https://help.duckduckgo.com/privacy/t/" target="_blank">[ref]</a>',
                    parent_id=node.node_id, incoming_edge_config=duckduckgo)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

mastodon_edge = {
    'color': {
        'color': '#419BDD'
    },
    'title': 'Mastodon Snowflake',
    'label': '❄'
}


def parse_mastodon_snowflake(unfurl, node):
    # Ref: https://github.com/tootsuite/mastodon/issues/1059
    #      https://github.com/tootsuite/mastodon/blob/master/lib/mastodon/snowflake.rb
    try:
        snowflake = int(node.value)
        seq_data = snowflake & 0xFF
        timestamp = (snowflake >> 16) 

    except Exception as e:
        print(e)
        return

    node.hover = 'Mastodon Snowflakes are time-based IDs similar to those of Twitter Snowflakes. ' \
                 '<a href="https://blog.twitter.com/engineering/en_us/a/2010/announcing-snowflake.html" ' \
                 'target="_blank">[ref]</a>'

    unfurl.add_to_queue(
        data_type='epoch-milliseconds', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
        hover='The first 16 bits value in a Mastodon Snowflake is a timestamp.',
        parent_id=node.node_id, incoming_edge_config=mastodon_edge)

    unfurl.add_to_queue(
         data_type='integer', key=None, value=seq_data, label=f'Sequence data: {seq_data}',
         hover="The 'sequence data' is intended to be unique within a given millisecond. It is a 2 byte value.",
         parent_id=node.node_id, incoming_edge_config=mastodon_edge)


def run(unfurl, node):

    # Known pattern from mastodon.social site
    if node.data_type == 'url.path.segment':
        # Mastodon by nature is federated and there are many, many domains hosting instances. This list was taken
        # from instances.social, sorted by popularity.
        mastodon_domains = ['pawoo.net', 'mastodon.social', 'mstdn.jp', 'switter.at', 'mastodon.cloud', 'baraag.net',
                            'sinblr.com', 'mastodon.xyz', 'mastodon.technology', 'mstdn.io', 'infosec.exchange']

        # Check if node is a child of a mastodon domain, is an integer, & timestamp would be between 2015-01 and 2030-01
        if any(mastodon_domain in unfurl.find_preceding_domain(node) for mastodon_domain in mastodon_domains) and \
                unfurl.check_if_int_between(node.value, 93065733734400000, 124089536413761540):
            parse_mastodon_snowflake(unfurl, node)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

bing_edge = {
    'color': {
        'color': '#228372'
    },
    'title': 'Bing-related Parsing Functions',
    'label': 'B'
}


def run(unfurl, node):
    if node.data_type == 'url.query.pair':
        if 'bing' in unfurl.find_preceding_domain(node):

            # Many of the corresponding names of these QSPs were found on bing.com/as/init?pt...
            # Their function is not completely understood.
            if node.key == 'pq':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'"Partial" Search Query: {node.value}',
                    hover='Partial search terms entered by the user; auto-complete or suggestions <br>'
                          'may have been used to reach the actual search terms (in <b>q</b>)',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'q':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Search Query: {node.value}',
                    hover='Terms used in the Bing search', parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'cp':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Cursor Position: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'cvid':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Conversation ID',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'sc':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Suggestion Count: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'sp':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Suggestion Position: {node.value}',
                    hover='An sp of "-1" indicates no suggestion was used for the search',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'qs':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Suggestion Type: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'qsc':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Preview Pane Suggestion Type: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'sk':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Skip Value: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'skc':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Preview Pane Skip Value: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'ghc':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Ghosting: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'ds':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Data Set: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'sid':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Session ID: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'qt':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Timestamp: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'ig':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Impression GUID: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'bq':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Base Query: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'nclid':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Hashed MUID: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'hgr':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Home Geographic Region: {node.value}',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)

            elif node.key == 'first':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Starting Result: {node.value}',
                    hover='Bing search by default shows 8 results per page; higher <br>'
                          '"first" values may indicate browsing more subsequent results pages.',
                    parent_id=node.node_id, incoming_edge_config=bing_edge)



# Copyright 2021 Ryan Benson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

brave_edge = {
    'color': {
        'color': '#fb542b'
    },
    'title': 'Brave Search-related Parsing Functions',
    'label': '🦁'
}


def run(unfurl, node):
    if 'search.brave' in unfurl.find_preceding_domain(node):
        if node.data_type == 'url.query.pair':
            if node.key == 'tf':
                tf_mappings = {
                    'pd': 'Past Day',
                    'pw': 'Past Week',
                    'pm': 'Past Month',
                    'py': 'Past Year'
                }

                tf_friendly_value = tf_mappings.get(node.value)
                if not tf_friendly_value:
                    m = re.fullmatch(r'(?P<start_date>\d{4}-\d{2}-\d{2})to(?P<end_date>\d{4}-\d{2}-\d{2})', node.value)
                    if m:
                        tf_friendly_value = f'{m.group("start_date")} to {m.group("end_date")}'
                    else:
                        tf_friendly_value = 'Unknown'

                unfurl.add_to_queue(
                    data_type='descriptor', key='Time Filter', value=tf_friendly_value,
                    hover='Time Filter', parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'q':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=f'Search Query: {node.value}',
                    hover='Terms used in the Brave search', parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'show_local':
                local_mapping = {'0': 'False', '1': 'True'}

                unfurl.add_to_queue(
                    data_type='descriptor', key='Localized Results',
                    value=f'{local_mapping.get(node.value, "Unknown")}',
                    hover='Location information is derived from user\'s IP address by default, but can be manually '
                          'changed. '
                          '<a href="https://search.brave.com/help/anonymous-local-results" target="_blank">[ref]</a>',
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'size':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Image Size Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == '_type':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Image Type Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'layout':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Image Layout Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'color':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Image Color Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'length':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Video Length Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

            elif node.key == 'resolution':
                unfurl.add_to_queue(
                    data_type='descriptor', key='Video Resolution Filter', value=node.value,
                    parent_id=node.node_id, incoming_edge_config=brave_edge)

        elif node.data_type == 'url.path':
            search_type_mappings = {
                '/images': 'Image Search',
                '/news': 'News Search',
                '/search': 'Web Search',
                '/videos': 'Video Search',
            }
            search_type = search_type_mappings.get(node.value)
            if search_type:
                unfurl.add_to_queue(
                    data_type='descriptor', key='Search Type', value=search_type,
                    hover='Brave Search Type', parent_id=node.node_id, incoming_edge_config=brave_edge)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import ulid

uuid_edge = {
    'color': {
        'color': 'orange'
    },
    'title': 'ULID Parsing Functions',
    'label': 'UL'
}


def run(unfurl, node):
    if not node.data_type.startswith('ulid'):
        # ULIDs are Base32 encoded and allow dashes to be added as desired for clarity
        # Ref:
        #   - https://github.com/ulid/spec
        #   - https://www.crockford.com/base32.html
        #
        # Filtering down to reasonable timeframes to reduce FPs:
        #   - 019AHCNC00SM9CSFQFXG3VC1FK <- 2015-01-01
        #   - 01JGFJJZ00TG242KAWHD959K7S <- 2025-01-01

        m = re.match(r'(?P<ulid>01[90A-HJ][A-HJKMNP-Z0-9]{23})', str(node.value).replace('-', ''))
        if m:
            u = m.group('ulid')
            unfurl.add_to_queue(
                data_type='ulid', key=None, value=u, label=f'ULID: {u}',
                hover='ULIDs are identifiers that are comprised of a timestamp and a random number. '
                      '<a href="https://github.com/ulid/spec" target="_blank">[ref]</a>',
                parent_id=node.node_id, incoming_edge_config=uuid_edge,
                extra_options={'widthConstraint': {'maximum': 300}})

    elif node.data_type == 'ulid':
        u = ulid.parse(node.value)
        timestamp = u.timestamp().int
        unfurl.add_to_queue(
            data_type='ulid-parsed', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
            parent_id=node.node_id, incoming_edge_config=uuid_edge)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import re
import zlib
from unfurl import utils

zip_edge = {
    'color': {
        'color': '#2C63FF'
    },
    'title': 'Compression-related Parsing Functions',
    'label': 'zip'
}

b64_zip_edge = {
    'color': {
        'color': '#2C63FF'
    },
    'title': 'Compression-related Parsing Functions',
    'label': 'b64+zip'
}


def run(unfurl, node):

    if not isinstance(node.value, str):
        return False

    if node.data_type == 'zlib':
        inflated_str = zlib.decompress(node.value)
        unfurl.add_to_queue(
            data_type='zlib-inflate', key=None, value=inflated_str,
            hover='This data was inflated using zlib',
            parent_id=node.node_id, incoming_edge_config=zip_edge)
        return

    if node.data_type in ('url.scheme', 'url.host', 'url.domain', 'url.tld'):
        return

    # This checks for base64 encoding, which is often used before compression. Initially, the base64 decoding was
    # in parse_base64.py, but the intermediary node seemed like not useful clutter. I moved it here and combined
    # the parser into b64+zlib
    if len(node.value) % 4 == 1:
        # A valid b64 string will not be this length
        return False

    urlsafe_b64_m = utils.urlsafe_b64_re.fullmatch(node.value)
    standard_b64_m = utils.standard_b64_re.fullmatch(node.value)
    long_int_m = utils.long_int_re.fullmatch(node.value)

    # Long integers pass the b64 regex, but we don't want those here.
    if long_int_m:
        return

    decoded = None
    padded_value = unfurl.add_b64_padding(node.value)
    if not padded_value:
        return

    if urlsafe_b64_m:
        decoded = base64.urlsafe_b64decode(unfurl.add_b64_padding(node.value))
    elif standard_b64_m:
        decoded = base64.b64decode(unfurl.add_b64_padding(node.value))

    if decoded == node.value or not decoded:
        return

    try:
        inflated_bytes = zlib.decompress(decoded)
    except:
        # If we can't inflate it, bail on this parser.
        return

    try:
        inflated_str = inflated_bytes.decode('ascii', errors='strict')
        if re.fullmatch(r'\w+', inflated_str):
            unfurl.add_to_queue(
                data_type='string', key=None, value=inflated_str,
                parent_id=node.node_id,
                hover='This data was base64-decoded, then zlib inflated',
                incoming_edge_config=b64_zip_edge)
            return
    except:
        # If we couldn't decode the inflated bytes as ASCII, that's ok; we'll
        # just show the raw inflated bytes below.
        pass

    unfurl.add_to_queue(
        data_type='bytes', key=None, value=inflated_bytes, parent_id=node.node_id,
        hover='This data was inflated using zlib', incoming_edge_config=b64_zip_edge)



# Copyright 2019 Dan Nemec
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib.parse
import warnings

try:
    from publicsuffix2 import PublicSuffixList
    psl = PublicSuffixList(idna=False)
except ImportError:
    warnings.warn("Unable to import the nodule 'publicsuffix2'. "
                  "Will be unable to parse domain names.")
    psl = None


urlparse_edge = {
    'color': {
        'color': '#4d4d4d'
    },
    'title': 'URL Parsing Functions',
    'label': 'u'
}


def run(unfurl, node):
    if node.data_type == 'url.domain':
        hits_in_known_lists = unfurl.search_known_domain_lists(node.value)
        for hit in hits_in_known_lists:
            cleaned_node_name = hit['name']
            cleaned_node_hover = hit['description']

            if cleaned_node_name.startswith('List of'):
                cleaned_node_name = f'Domain is on list {cleaned_node_name[5:]}'

            event_desc_message_prefix = 'Event contains one or more entries of known'
            if cleaned_node_hover.startswith(event_desc_message_prefix):
                cleaned_node_hover = f'Domain is on list of {cleaned_node_hover[len(event_desc_message_prefix):]}'

            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=cleaned_node_name, hover=cleaned_node_hover,
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    if psl is None or node.data_type != 'url.hostname' or not isinstance(node.value, str):
        return

    full_domain = node.value
    if full_domain.startswith('xn--') or '.xn--' in full_domain:
        # punycoded domain
        full_domain = full_domain.encode('utf8').decode('idna')
    if '%' in full_domain:
        # percent-encoded domain
        full_domain = urllib.parse.unquote(full_domain)

    domain = psl.get_sld(full_domain)
    if domain is not None:
        if len(domain) != len(full_domain):
            subdomain = full_domain[:-len(domain)-1]
            unfurl.add_to_queue(
                data_type='url.subdomain', key='Subdomain', value=subdomain,
                hover='This is the <b>sub-domain</b> part of the domain or netloc.',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)
        unfurl.add_to_queue(
            data_type='url.domain', key='Domain Name', value=domain,
            hover='This is the base, registrable, part of the domain or netloc',
            parent_id=node.node_id, incoming_edge_config=urlparse_edge)
        
    tld = psl.get_tld(full_domain)
    if tld is not None:
        unfurl.add_to_queue(
            data_type='url.tld', key='TLD', value=tld,
            hover='This is the <b>Top Level Domain</b>, or TLD, for the domain or netloc',
            parent_id=node.node_id, incoming_edge_config=urlparse_edge)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google_search.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
  name='google_search.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x13google_search.proto\"\xd1\x03\n\x03Ved\x12\x12\n\nlink_index\x18\x01 \x01(\x05\x12\x11\n\tlink_type\x18\x02 \x01(\x05\x12\n\n\x02v3\x18\x03 \x01(\x05\x12\n\n\x02v4\x18\x04 \x01(\x05\x12\x1b\n\x13sub_result_position\x18\x05 \x01(\x05\x12\x17\n\x0fresult_position\x18\x06 \x01(\x05\x12\x15\n\rresults_start\x18\x07 \x01(\x05\x12\n\n\x02v8\x18\x08 \x01(\x05\x12\n\n\x02v9\x18\t \x01(\x05\x12\x0b\n\x03v10\x18\n \x01(\x05\x12\x0b\n\x03v11\x18\x0b \x01(\x05\x12\x0b\n\x03v12\x18\x0c \x01(\x05\x12%\n\tv13_Outer\x18\r \x01(\x0b\x32\x12.Ved.v13_Outer_Msg\x12\x0b\n\x03v14\x18\x0e \x01(\x05\x12\x19\n\x03v15\x18\x0f \x01(\x0b\x32\x0c.Ved.v15_Msg\x1a\x86\x01\n\rv13_Outer_Msg\x12\x33\n\tv13_Inner\x18\x01 \x01(\x0b\x32 .Ved.v13_Outer_Msg.v13_Inner_Msg\x1a@\n\rv13_Inner_Msg\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05v13_2\x18\x02 \x01(\x07\x12\r\n\x05v13_3\x18\x03 \x01(\x07\x1a\'\n\x07v15_Msg\x12\r\n\x05v15_1\x18\x01 \x01(\x05\x12\r\n\x05v15_2\x18\x02 \x01(\x05\x62\x06proto3'
)


_VED_V13_OUTER_MSG_V13_INNER_MSG = _descriptor.Descriptor(
  name='v13_Inner_Msg',
  full_name='Ved.v13_Outer_Msg.v13_Inner_Msg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='Ved.v13_Outer_Msg.v13_Inner_Msg.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v13_2', full_name='Ved.v13_Outer_Msg.v13_Inner_Msg.v13_2', index=1,
      number=2, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v13_3', full_name='Ved.v13_Outer_Msg.v13_Inner_Msg.v13_3', index=2,
      number=3, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=384,
  serialized_end=448,
)

_VED_V13_OUTER_MSG = _descriptor.Descriptor(
  name='v13_Outer_Msg',
  full_name='Ved.v13_Outer_Msg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='v13_Inner', full_name='Ved.v13_Outer_Msg.v13_Inner', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_VED_V13_OUTER_MSG_V13_INNER_MSG, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=314,
  serialized_end=448,
)

_VED_V15_MSG = _descriptor.Descriptor(
  name='v15_Msg',
  full_name='Ved.v15_Msg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='v15_1', full_name='Ved.v15_Msg.v15_1', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v15_2', full_name='Ved.v15_Msg.v15_2', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=450,
  serialized_end=489,
)

_VED = _descriptor.Descriptor(
  name='Ved',
  full_name='Ved',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='link_index', full_name='Ved.link_index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link_type', full_name='Ved.link_type', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v3', full_name='Ved.v3', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v4', full_name='Ved.v4', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sub_result_position', full_name='Ved.sub_result_position', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_position', full_name='Ved.result_position', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='results_start', full_name='Ved.results_start', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v8', full_name='Ved.v8', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v9', full_name='Ved.v9', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v10', full_name='Ved.v10', index=9,
      number=10, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v11', full_name='Ved.v11', index=10,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v12', full_name='Ved.v12', index=11,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v13_Outer', full_name='Ved.v13_Outer', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v14', full_name='Ved.v14', index=13,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v15', full_name='Ved.v15', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_VED_V13_OUTER_MSG, _VED_V15_MSG, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24,
  serialized_end=489,
)

_VED_V13_OUTER_MSG_V13_INNER_MSG.containing_type = _VED_V13_OUTER_MSG
_VED_V13_OUTER_MSG.fields_by_name['v13_Inner'].message_type = _VED_V13_OUTER_MSG_V13_INNER_MSG
_VED_V13_OUTER_MSG.containing_type = _VED
_VED_V15_MSG.containing_type = _VED
_VED.fields_by_name['v13_Outer'].message_type = _VED_V13_OUTER_MSG
_VED.fields_by_name['v15'].message_type = _VED_V15_MSG
DESCRIPTOR.message_types_by_name['Ved'] = _VED
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Ved = _reflection.GeneratedProtocolMessageType('Ved', (_message.Message,), {

  'v13_Outer_Msg' : _reflection.GeneratedProtocolMessageType('v13_Outer_Msg', (_message.Message,), {

    'v13_Inner_Msg' : _reflection.GeneratedProtocolMessageType('v13_Inner_Msg', (_message.Message,), {
      'DESCRIPTOR' : _VED_V13_OUTER_MSG_V13_INNER_MSG,
      '__module__' : 'google_search_pb2'
      # @@protoc_insertion_point(class_scope:Ved.v13_Outer_Msg.v13_Inner_Msg)
      })
    ,
    'DESCRIPTOR' : _VED_V13_OUTER_MSG,
    '__module__' : 'google_search_pb2'
    # @@protoc_insertion_point(class_scope:Ved.v13_Outer_Msg)
    })
  ,

  'v15_Msg' : _reflection.GeneratedProtocolMessageType('v15_Msg', (_message.Message,), {
    'DESCRIPTOR' : _VED_V15_MSG,
    '__module__' : 'google_search_pb2'
    # @@protoc_insertion_point(class_scope:Ved.v15_Msg)
    })
  ,
  'DESCRIPTOR' : _VED,
  '__module__' : 'google_search_pb2'
  # @@protoc_insertion_point(class_scope:Ved)
  })
_sym_db.RegisterMessage(Ved)
_sym_db.RegisterMessage(Ved.v13_Outer_Msg)
_sym_db.RegisterMessage(Ved.v13_Outer_Msg.v13_Inner_Msg)
_sym_db.RegisterMessage(Ved.v15_Msg)


# @@protoc_insertion_point(module_scope)



from setuptools import setup, find_packages

setup(
  name='dfir-unfurl',
  packages=find_packages(),
  include_package_data=True,
  scripts=['unfurl_app.py', 'unfurl_cli.py'],
  version='20220200',
  description='Unfurl takes a URL and expands ("unfurls") it into a directed graph',
  url='https://github.com/obsidianforensics/unfurl',
  author='Ryan Benson',
  author_email='ryan@dfir.blog',
  license='Apache',
  keywords=['unfurl', 'forensics', 'dfir', 'reverse-engineering', 'security'],
  classifiers=[],
  install_requires=[
    'flask',
    'flask_cors',
    'maclookup',
    'networkx',
    'protobuf',
    'publicsuffix2',
    'pycountry',
    'pymispwarninglists',
    'Requests',
    'torf',
    'ulid-py'
  ]
)




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
from unfurl import utils

hash_edge = {
    'color': {
        'color': '#4A93AE'
    },
    'title': 'Hash Identification Functions',
    'label': '#'
}

hash_lookup_edge = {
    'color': {
        'color': '#4A93AE'
    },
    'title': 'Hash Lookup Functions',
    'label': '#'
}


def nitrxgen_md5_lookup(value):
    response = requests.get(f'https://www.nitrxgen.net/md5db/{value}', verify=False).text

    if response:
        return response
    else:
        return False


def virustotal_lookup(unfurl, hash_value):

    response = requests.get(f'https://www.virustotal.com/api/v3/files/{hash_value}',
                            headers={'x-apikey': unfurl.api_keys.get('virustotal')})

    if response.status_code == 200:
        try:
            result = response.json()
            return result['data']['attributes']
        except:
            return False


def decode_cisco_type_7(encoded_text):
    cisco_constant = b"dsfd;kfoA,.iyewrkldJKDHSUBsgvca69834ncxv9873254k;fg87"
    try:
        salt = int(encoded_text[0:2])
    except ValueError:
        # Valid salts should be ints; if not, move on.
        return

    try:
        encoded = bytearray.fromhex(encoded_text[2:])
    except ValueError:
        # Not valid Type 7 encoded then; exit
        return

    plaintext = ''
    for i in range(0, len(encoded)):
        j = (i + salt) % 53
        p = encoded[i] ^ cisco_constant[j]
        plaintext += chr(p)

    # If the result isn't readable as ASCII, call it a false positive and move on without adding a node.
    try:
        _ = plaintext.encode('ascii', errors='strict')
    except UnicodeEncodeError:
        return

    return plaintext


def run(unfurl, node):

    if node.data_type.startswith('uuid'):
        return

    if node.data_type.startswith('hash'):
        if node.data_type == 'hash.md5' and unfurl.remote_lookups:
            hash_plaintext = nitrxgen_md5_lookup(node.value)

            if hash_plaintext:
                unfurl.add_to_queue(
                    data_type=f'text', key='Plaintext',
                    value=hash_plaintext,
                    hover='Queried Nitrxgen database of MD5 hashes and found a matching plaintext value',
                    parent_id=node.node_id, incoming_edge_config=hash_lookup_edge)

        if node.data_type in ('hash.md5', 'hash.sha-1', 'hash.sha-256') and unfurl.remote_lookups:
            vt_results = virustotal_lookup(unfurl, node.value)
            if vt_results:
                label_text = 'Hash found on VirusTotal'
                if vt_results.get("type_description"):
                    label_text += f'\nFile Type: {vt_results.get("type_description")};'

                if vt_results.get("meaningful_name"):
                    label_text += f'\nName: {vt_results.get("meaningful_name")};'

                if vt_results.get("reputation"):
                    label_text += f'\nReputation: {vt_results.get("reputation")};'

                unfurl.add_to_queue(
                    data_type=f'text', key='Hash found on VirusTotal',
                    value=None, label=label_text,
                    hover='Queried VirusTotal with the hash value and found a match.',
                    parent_id=node.node_id, incoming_edge_config=hash_lookup_edge)

    else:
        if not isinstance(node.value, str):
            return

        # Filter for values that are only hex chars (A-F,0-9) and contains both a letter and number.
        # This could conceivably filter out valid hashes, but will filter out many more invalid values.
        if not (utils.hex_re.fullmatch(node.value) and
                utils.digits_re.search(node.value) and utils.letters_re.search(node.value)):
            return

        # Cisco "Type 7" password encoding is very flexible, so detecting it is very false positive prone
        # as it isn't a fixed length. However, decoding it is easy, so Unfurl will only "detect" something as
        # using this encoding type if it can also decode it (as a method of verifying it).
        # Ref: https://passlib.readthedocs.io/en/stable/lib/passlib.hash.cisco_type7.html
        cisco_type_7_m = utils.cisco_7_re.fullmatch(node.value)
        if cisco_type_7_m:
            cisco_type_7_plaintext = decode_cisco_type_7(node.value)
            if cisco_type_7_plaintext:
                unfurl.add_to_queue(
                    data_type=f'text', key=f'Cisco "Type 7" encoding', value=cisco_type_7_plaintext,
                    label=f'Cisco "Type 7" encoding; plaintext is "{cisco_type_7_plaintext}"',
                    hover='Cisco "Type 7" password encoding is based<br> on XOR and is easily reversible '
                          '[<a hre="https://passlib.readthedocs.io/en/stable/lib/passlib.hash.cisco_type7.html">'
                          'ref</a>].', parent_id=node.node_id, incoming_edge_config=hash_edge)
            return

        if len(node.value) == 32 and node.value[12] == '4':
            # UUIDv4 is very common and it's the same length as an MD5 hash. This might filter out some legitimate
            # MD5 hashes, but it will filter out many more UUIDs. I think the tradeoff is worth it for Unfurl.
            return

        hash_name, hash_hover, new_node_value = None, None, None

        if len(node.value) == 32:
            hash_name = 'MD5'
            hash_hover = f'This is potentially a <b>{hash_name}</b> hash <br>(based on length and character set).'

        if len(node.value) == 40:
            hash_name = 'SHA-1'
            hash_hover = f'This is potentially a <b>{hash_name}</b> hash <br>(based on length and character set).'

        if len(node.value) == 64:
            hash_name = 'SHA-256'
            hash_hover = f'This is potentially a <b>{hash_name}</b> hash <br>(based on length and character set).'

        if len(node.value) == 128:
            hash_name = 'SHA-512'
            hash_hover = f'This is potentially a <b>{hash_name}</b> hash <br>(based on length and character set).'

        if hash_name in ('MD5', 'SHA-1', 'SHA-256'):
            # Pass through the values of three common file hashes for further analysis; don't send on the
            # other types to avoid duplicate processing.
            new_node_value = node.value

        if hash_name:
            unfurl.add_to_queue(
                data_type=f'hash.{hash_name.lower()}', key=f'{hash_name} Hash',
                value=new_node_value, label=f'Potential {hash_name} hash',
                hover=hash_hover, parent_id=node.node_id, incoming_edge_config=hash_edge)




#!/usr/bin/env python3

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
from unfurl import core

config = configparser.ConfigParser()
config.read('unfurl.ini')

unfurl_host = 'localhost'
unfurl_port = '5000'
unfurl_debug = 'True'
remote_lookups = False

if config.has_section('UNFURL_APP'):
    unfurl_host = config['UNFURL_APP'].get('host')
    unfurl_port = config['UNFURL_APP'].get('port')
    unfurl_debug = config['UNFURL_APP'].get('debug')
    try:
        remote_lookups = config['UNFURL_APP'].getboolean('remote_lookups')
    # If we can't interpret it as a boolean, fail "safe" to not allowing lookups
    except ValueError:
        remote_lookups = False


if __name__ == '__main__':
    core.UnfurlApp(
        unfurl_debug=unfurl_debug,
        unfurl_host=unfurl_host,
        unfurl_port=unfurl_port,
        remote_lookups=remote_lookups)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from unfurl import utils

b64_edge = {
    'color': {
        'color': '#2C63FF'
    },
    'title': 'Base64 Parsing Functions',
    'label': 'b64'
}


def run(unfurl, node):

    if not isinstance(node.value, str):
        return False

    if len(node.value) % 4 == 1:
        # A valid b64 string will not be this length
        return False

    urlsafe_b64_m = utils.urlsafe_b64_re.fullmatch(node.value)
    standard_b64_m = utils.standard_b64_re.fullmatch(node.value)
    long_int_m = utils.long_int_re.fullmatch(node.value)

    # Long integers pass the b64 regex, but we don't want those here.
    if long_int_m:
        return

    decoded = None
    padded_value = unfurl.add_b64_padding(node.value)
    if not padded_value:
        return

    if urlsafe_b64_m:
        decoded = base64.urlsafe_b64decode(padded_value)
    elif standard_b64_m:
        decoded = base64.b64decode(padded_value)

    if decoded == node.value or not decoded:
        return

    try:
        # This limits the plugin to only decoding ASCII string that were base64
        # encoded. Obviously other things could be encoded, but it's a start.
        str_decoded = decoded.decode('ascii', errors='strict')

    # This will happen a lot with things that aren't really b64 encoded, or
    # with things that are b64-encoded, but the results are not ASCII
    # (like gzip or protobufs).
    except UnicodeDecodeError:
        # Show the resulting bytes from base64 inflating. Disabled for now,
        # as it's too noisy.
        # unfurl.add_to_queue(data_type='bytes', key=None, value=decoded,
        #                parent_id=node.node_id, incoming_edge_config=b64_edge)
        return

    unfurl.add_to_queue(data_type='string', key=None, value=str_decoded,
                        parent_id=node.node_id, incoming_edge_config=b64_edge)



__all__ = [
    "parse_base64",
    "parse_bing",
    "parse_brave",
    "parse_compressed",
    "parse_discord",
    "parse_domain",
    "parse_dropbox",
    "parse_duckduckgo",
    "parse_google",
    "parse_hash",
    "parse_json",
    "parse_ksuid",
    "parse_mac_addr",
    "parse_magnet",
    "parse_mastodon",
    "parse_mailto",
    "parse_metasploit",
    "parse_protobuf",
    "parse_shortlink",
    "parse_sonyflake",
    "parse_tiktok",
    "parse_timestamp",
    "parse_twitter",
    "parse_ulid",
    "parse_url",
    "parse_uuid",
    "parse_youtube",
    "parse_yahoo"
]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re
from unfurl import utils

timestamp_edge = {
    'color': {
        'color': 'blue'
    },
    'title': 'Date & Time Parsing Functions',
    'label': '🕓'
}


def trim_zero_fractional_seconds(timestamp_string, number_to_trim):
    """Timestamp formats have different levels of precision; trim off extra 0s.

    Different formats may have less precision that the microseconds datetime returns.
    Trim off the appropriate number of trailing zeros from a value to not add extra,
    incorrect precision to it.

    """
    if re.search(rf'\.\d{{{6 - number_to_trim}}}0{{{number_to_trim}}}$', timestamp_string):
        return timestamp_string[:-number_to_trim]
    return timestamp_string


def decode_epoch_seconds(seconds):
    """Decode a numeric timestamp in Epoch seconds format to a human-readable timestamp.

    An Epoch timestamp (1-10 digits) is an integer that counts the number of seconds since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 0
      2015: 1420070400
      2025: 1735689600
      2030: 1900000000

    """
    return datetime.datetime.utcfromtimestamp(float(seconds)), 'Epoch seconds'


def decode_epoch_centiseconds(centiseconds):
    """Decode a numeric timestamp in Epoch centiseconds (10 ms) format to a human-readable timestamp.

    An Epoch centisecond timestamp (1-12 digits) is an integer that counts the number of centiseconds (10 ms)
    since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 0
      2015: 142007040000
      2025: 173568960000
      2030: 190000000000

    """
    # Trim off the 4 trailing 0s (don't add precision that wasn't in the timestamp)
    converted_ts = trim_zero_fractional_seconds(
        str(datetime.datetime.utcfromtimestamp(float(centiseconds) / 100)), 4)
    return converted_ts, 'Epoch centiseconds'


def decode_epoch_milliseconds(milliseconds):
    """Decode a numeric timestamp in Epoch milliseconds format to a human-readable timestamp.

    An Epoch millisecond timestamp (1-13 digits) is an integer that counts the number of milliseconds since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 0
      2015: 1420070400000
      2025: 1735689600000
      2030: 1900000000000

    """
    converted_dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=float(milliseconds))
    # Trim off the 3 trailing 0s (don't add precision that wasn't in the timestamp)
    converted_ts = trim_zero_fractional_seconds(str(converted_dt), 3)
    return converted_ts, 'Epoch milliseconds'


def decode_epoch_ten_microseconds(ten_microseconds):
    """Decode a numeric timestamp in Epoch ten-millisecond increments to a human-readable timestamp.

    An Epoch ten-microsecond increments timestamp (1-15 digits) is an integer that counts the number of ten-microsecond
    increments since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 0
      2015: 142007040000000
      2025: 173568960000000
      2030: 190000000000000

    """
    # Trim off the trailing 0 (don't add precision that wasn't in the timestamp)
    converted_ts = trim_zero_fractional_seconds(
        str(datetime.datetime.utcfromtimestamp(float(ten_microseconds) / 100000)), 1)
    return converted_ts, 'Epoch ten-microsecond increments'


def decode_epoch_microseconds(microseconds):
    """Decode a numeric timestamp in Epoch microseconds format to a human-readable timestamp.

    An Epoch millisecond timestamp (1-16 digits) is an integer that counts the number of milliseconds since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 0
      2015: 1420070400000000
      2025: 1735689600000000
      2030: 1900000000000000

    """
    converted_ts = str(datetime.datetime.utcfromtimestamp(float(microseconds) / 1000000))
    return converted_ts, 'Epoch microseconds'


def decode_webkit(microseconds):
    """Decode a numeric timestamp in Webkit format to a human-readable timestamp.

    A Webkit timestamp (17 digits) is an integer that counts the number of microseconds since 12:00AM Jan 1 1601 UTC.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 11644473600000000
      2015: 13064544000000000
      2025: 13380163200000000

    """
    return datetime.datetime.utcfromtimestamp((float(microseconds) / 1000000) - 11644473600), 'Webkit'


def decode_windows_filetime(intervals):
    """Decode a numeric timestamp in Windows FileTime format to a human-readable timestamp.

    A Windows FileTime timestamp (18 digits) is a 64-bit value that represents the number of 100-nanosecond intervals
    since 12:00AM Jan 1 1601 UTC.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 116444736000000000
      2015: 130645440000000000
      2025: 133801632000000000
      2065: 146424672000000000

    """
    return datetime.datetime.utcfromtimestamp((float(intervals) / 10000000) - 11644473600), 'Windows FileTime'


def decode_datetime_ticks(ticks):
    """Decode a numeric timestamp in .Net/C# DateTime ticks format to a human-readable timestamp.

    A .Net/C# DateTime ticks timestamp (18 digits) is the number of 100-nanosecond intervals that have elapsed since
    12:00:00 midnight, January 1, 0001 (0:00:00 UTC on January 1, 0001, in the Gregorian calendar), which represents
    DateTime.MinValue. It does not include the number of ticks that are attributable to leap seconds.

    A single tick represents one hundred nanoseconds or one ten-millionth of a second. There are 10,000 ticks in a
    millisecond, or 10 million ticks in a second.

    (^ from https://docs.microsoft.com/en-us/dotnet/api/system.datetime.ticks?view=netframework-4.8)

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 621355968000000000
      2015: 635556672000000000
      2025: 638712864000000000
      2038: 642815136000000000

    """
    seconds = (ticks - 621355968000000000) / 10000000
    return (datetime.datetime.fromtimestamp(seconds)), 'DateTime ticks'


def decode_mac_absolute_time(seconds):
    """Decode a numeric timestamp in Mac Absolute Time format to a human-readable timestamp.

    A Mac Absolute Time timestamp (9 digits typically) is the number of seconds since 2001-01-01. This time format is
    also known as CFAbsoluteTime, Core Data timestamp, or Cocoa Core Data timestamp. Negative values are allowed and
    denote timestamps before the reference date.

    ref: https://developer.apple.com/documentation/corefoundation/cfabsolutetime

    Useful values for ranges (all Jan-1 00:00:00):
      1970: -978307200
      2015: 441763200
      2025: 757382400
      2035: 1072915200

    """
    return datetime.datetime.utcfromtimestamp(float(seconds)+978307200), 'Mac Absolute Time / Cocoa'


def decode_epoch_hex(seconds):
    """Decode a hex string (big endian) of an Epoch seconds integer to a human-readable timestamp.

    An Epoch timestamp (1-10 digits) is an integer that counts the number of seconds since Jan 1 1970.

    Useful values for ranges (all Jan-1 00:00:00):
      2015: 54A48E00
      2025: 67748580
      2030: 713FB300

    """
    timestamp, _ = decode_epoch_seconds(int(seconds, 16))
    return timestamp, 'Epoch seconds (hex)'


def decode_windows_filetime_hex(intervals):
    """Decode a hex timestamp in Windows FileTime format to a human-readable timestamp.

    A Windows FileTime timestamp (18 digits) is a 64-bit value that represents the number of 100-nanosecond intervals
    since 12:00AM Jan 1 1601 UTC.

    Useful values for ranges (all Jan-1 00:00:00):
      1970: 19DB1DED53E8000
      2015: 1D02555E2B98000
      2025: 1DB5BE019BA4000
      2065: 2083476A0E9C000

    """
    int_right = int(intervals, 16)
    timestamp, _ = decode_windows_filetime(int_right)
    return timestamp, 'Windows FileTime (hex)'


def run(unfurl, node):
    new_timestamp = (None, 'unknown')

    # If the node is explicitly classified as a raw timestamp, use that
    # type for the conversion
    if node.data_type == 'epoch-seconds':
        new_timestamp = decode_epoch_seconds(node.value)

    elif node.data_type == 'epoch-centiseconds':
        new_timestamp = decode_epoch_centiseconds(node.value)

    elif node.data_type == 'epoch-milliseconds':
        new_timestamp = decode_epoch_milliseconds(node.value)

    elif node.data_type == 'epoch-ten-microseconds':
        new_timestamp = decode_epoch_ten_microseconds(node.value)

    elif node.data_type == 'epoch-microseconds':
        new_timestamp = decode_epoch_microseconds(node.value)

    elif node.data_type == 'windows-filetime':
        new_timestamp = decode_windows_filetime(node.value)

    elif node.data_type == 'webkit':
        new_timestamp = decode_webkit(node.value)

    elif node.data_type == 'datetime-ticks':
        new_timestamp = decode_datetime_ticks(node.value)

    elif node.data_type == 'mac-absolute-time':
        new_timestamp = decode_mac_absolute_time(node.value)

    elif node.data_type == 'epoch-hex-seconds':
        new_timestamp = decode_epoch_hex(node.value)

    # Otherwise, examine the value of the node and see if we can detect a reasonable timestamp
    else:
        # There are some known cases where we want to suppress a timestamp conversion; put them here.
        if node.data_type == 'google.ei':
            return

        matches_digits = utils.digits_re.fullmatch(str(node.value))
        matches_float = utils.float_re.fullmatch(str(node.value))
        matches_hex = utils.hex_re.fullmatch(str(node.value))

        if matches_digits:
            timestamp = int(node.value)

            # Windows FileTime (18 digits)
            if 130645440000000000 <= timestamp <= 133801632000000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_windows_filetime(timestamp)

            # .Net/C# DateTime ticks (18 digits)
            elif 635556672000000000 <= timestamp <= 638712864000000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_datetime_ticks(timestamp)

            # WebKit (17 digits)
            elif 13064544000000000 <= timestamp <= 13380163200000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_webkit(timestamp)

            # Epoch microseconds (16 digits)
            elif 1420070400000000 <= timestamp <= 1735689600000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_microseconds(timestamp)

            # Epoch ten microsecond increments (15 digits)
            elif 142007040000000 <= timestamp <= 173568960000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_microseconds(timestamp)

            # Epoch milliseconds (13 digits)
            elif 1420070400000 <= timestamp <= 1735689600000:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_milliseconds(timestamp)

            # Epoch seconds (10 digits)
            elif 1420070400 <= timestamp <= 1735689600:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_seconds(timestamp)

            # Mac Absolute Time (9 digits)
            elif 441763200 <= timestamp <= 757382400:  # 2015 <= ts <= 2025
                new_timestamp = decode_mac_absolute_time(timestamp)

        elif matches_float:
            timestamp = float(node.value)

            # Epoch seconds (10 digits)
            if 1420070400.0 <= timestamp <= 1735689600.0:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_seconds(timestamp)

            # Mac Absolute Time (9 digits)
            elif 441763200.0 <= timestamp <= 757382400.0:  # 2015 <= ts <= 2025
                new_timestamp = decode_mac_absolute_time(timestamp)

        elif matches_hex:
            timestamp = node.value.replace(':', '')

            # Epoch hex seconds (8 hex chars)
            if 1420070400 <= int(timestamp, 16) <= 1735689600:  # 2015 <= ts <= 2025
                new_timestamp = decode_epoch_hex(timestamp)

            # Windows FileTime hex (16 hex digits)
            elif 130645440000000000 <= int(timestamp, 16) <= 133801632000000000:  # 2015 <= ts <= 2025
                new_timestamp = decode_windows_filetime_hex(timestamp)

    if new_timestamp != (None, 'unknown'):
        unfurl.add_to_queue(
            data_type=new_timestamp[1], key=None, value=new_timestamp[0],
            hover=f'Converted as {new_timestamp[1]}', parent_id=node.node_id,
            incoming_edge_config=timestamp_edge)



""" Functions for encoding and decoding fixed size structs """
import struct
import binascii
import six
from unfurl.lib.blackboxprotobuf.lib.exceptions import DecoderException, EncoderException

# Generic functions for encoding/decoding structs based on the "struct" format
def encode_struct(fmt, value):
    """Generic method for encoding arbitrary python "struct" values"""
    try:
        return struct.pack(fmt, value)
    except struct.error as exc:
        six.raise_from(
            EncoderException(
                "Error encoding value %r with format string %s" % (value, fmt)
            ),
            exc,
        )


def decode_struct(fmt, buf, pos):
    """Generic method for decoding arbitrary python "struct" values"""
    new_pos = pos + struct.calcsize(fmt)
    try:
        return struct.unpack(fmt, buf[pos:new_pos])[0], new_pos
    except struct.error as exc:
        six.raise_from(
            DecoderException(
                "Error deocding format string %s from bytes: %s"
                % (fmt, binascii.hexlify(buf[pos:new_pos]))
            ),
            exc,
        )


_fixed32_fmt = "<I"


def encode_fixed32(value):
    """Encode a single 32 bit fixed-size value"""
    return encode_struct(_fixed32_fmt, value)


def decode_fixed32(buf, pos):
    """Decode a single 32 bit fixed-size value"""
    return decode_struct(_fixed32_fmt, buf, pos)


_sfixed32_fmt = "<i"


def encode_sfixed32(value):
    """Encode a single signed 32 bit fixed-size value"""
    return encode_struct(_sfixed32_fmt, value)


def decode_sfixed32(buf, pos):
    """Decode a single signed 32 bit fixed-size value"""
    return decode_struct(_sfixed32_fmt, buf, pos)


_float_fmt = "<f"


def encode_float(value):
    """Encode a single 32 bit floating point value"""
    return encode_struct(_float_fmt, value)


def decode_float(buf, pos):
    """Decode a single 32 bit floating point value"""
    return decode_struct(_float_fmt, buf, pos)


_fixed64_fmt = "<Q"


def encode_fixed64(value):
    """Encode a single 64 bit fixed-size value"""
    return encode_struct(_fixed64_fmt, value)


def decode_fixed64(buf, pos):
    """Decode a single 64 bit fixed-size value"""
    return decode_struct(_fixed64_fmt, buf, pos)


_sfixed64_fmt = "<q"


def encode_sfixed64(value):
    """Encode a single signed 64 bit fixed-size value"""
    return encode_struct(_sfixed64_fmt, value)


def decode_sfixed64(buf, pos):
    """Decode a single signed 64 bit fixed-size value"""
    return decode_struct(_sfixed64_fmt, buf, pos)


_double_fmt = "<d"


def encode_double(value):
    """Encode a single 64 bit floating point value"""
    return encode_struct(_double_fmt, value)


def decode_double(buf, pos):
    """Decode a single 64 bit floating point value"""
    return decode_struct(_double_fmt, buf, pos)
